;; Generated package description from lem.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "lem" "0.24.0.20250806.92416" "A lemmy client" '((emacs "29.1") (fedi "0.2") (markdown-mode "2.5")) :commit "a0f4fa89fe73dfe7412f5d25d6e0619abf8cff14" :authors '(("martian hiatus" . "mousebot@disroot.org")) :maintainer '("martian hiatus" . "mousebot@disroot.org") :keywords '("multimedia" "comm" "web" "fediverse") :url "https://codeberg.org/martianh/lem.el")
