;;; Directory Local Variables            -*- no-byte-compile: t -*-
;;; For more information see (info "(emacs) Directory Variables")

((emacs-lisp-mode . ((comment-fill-column . 76)
                     (package-lint-main-file . "lem.el"))))
